import shutil
from pathlib import Path
import h5py
import numpy as np
import pandas as pd
import skimage

from extensions.chaeo.accessors import MonoPatchStackFromFile
from extensions.chaeo.models import generate_ilastik_object_classifier, PatchStackObjectClassifier
from model_server.accessors import GenericImageDataAccessor, write_accessor_data_to_file


def compare_object_maps(truth: GenericImageDataAccessor, inferred: GenericImageDataAccessor) -> pd.DataFrame:
    """
    Compare two object maps to assess classification results
    :param truth: t-stack of truth objects
    :param inferred: t-stack of inferred objects, presumably with same segmentation boundaries as truth
    :return: DataFrame comparing results for each frame in truth and inferred stacks
    """
    assert truth.shape == inferred.shape
    assert np.all((truth.data == 0) == (inferred.data == 0))
    assert inferred.chroma == 1

    labels = []
    for zi in range(0, inferred.nz):
        inf_img = inferred.data[:, :, :, zi]

        unique = np.unique(inf_img)
        assert unique[0] == 0

        dd = {'zi': zi, 'truth_label': np.unique(truth.data[:, :, :, zi])[1], 'multiples': False}

        if len(unique) == 1:  # no object in frame
            dd['inferred_label'] = unique[0]
        elif len(unique) > 2:  # multiple objects in frame, so mask out all but largest
            ob_id = skimage.measure.label(inf_img)
            pr = skimage.measure.regionprops_table(ob_id, properties=['label', 'area'])
            mask = inf_img == pr['label'][pr['area'].argmax()]
            dd['inferred_label'] = np.unique(mask * inf_img)[-1]  # occasionally no object in frame
            dd['multiples'] = True
        else:  # exactly one unique object class in frame
            dd['inferred_label'] = unique[1]
        labels.append(dd)
    return pd.DataFrame(labels)

def infer_and_compare(classifier: PatchStackObjectClassifier, prefix, raw, mask, labels):
    result_acc, _ = classifier.infer(raw, mask)
    write_accessor_data_to_file(root / f'zstack_train_result.tif', result_acc)

    # write comparison tables
    df_comp = compare_object_maps(labels, result_acc)
    df_comp.to_csv(root / f'compare_{prefix}_result.csv', index=False)
    print(f'Generated ilastik project {classifier_file.name}')
    print('Truth and inferred labels match?')
    print(pd.value_counts(df_comp['truth_label'] == df_comp['inferred_label']))


if __name__ == '__main__':
    root = Path('c:/Users/rhodes/projects/proj0011-plankton-seg/exp0009/output/labeled_patches-20231030-0002')
    template_ilp = Path('c:/Users/rhodes/projects/proj0011-plankton-seg/exp0014/template_obj.ilp')

    df_labels = pd.read_csv(root / 'labels_key.csv')
    label_names = list(df_labels.sort_values('annotation_class_id').annotation_class.unique())
    label_names[0] = 'none'
    assert len(label_names) >= 2

    # auto-populate an object classifier
    classifier_file = generate_ilastik_object_classifier(
        template_ilp,
        root / 'new_auto_obj.ilp',
        MonoPatchStackFromFile(root / 'zstack_train_raw.tif'),
        MonoPatchStackFromFile(root / 'zstack_train_mask.tif'),
        MonoPatchStackFromFile(root / 'zstack_train_label.tif'),
        label_names,
    )
    classifier = PatchStackObjectClassifier({'project_file': classifier_file})

    # verify self-consistency of training set
    infer_and_compare(
        classifier,
        'train',
        MonoPatchStackFromFile(root / 'zstack_train_raw.tif'),
        MonoPatchStackFromFile(root / 'zstack_train_mask.tif'),
        MonoPatchStackFromFile(root / 'zstack_train_label.tif')
    )

    # run test set
    infer_and_compare(
        classifier,
        'test',
        MonoPatchStackFromFile(root / 'zstack_test_raw.tif'),
        MonoPatchStackFromFile(root / 'zstack_test_mask.tif'),
        MonoPatchStackFromFile(root / 'zstack_test_label.tif'),
    )

